import os
from openjudge import config, tools
from subprocess import (run, PIPE, TimeoutExpired)
from multiprocessing import Queue
from threading import Thread


def __thread_worker__():
    """
    Keep checking the job queue for a job and check it.
    Place it in contest data when it is done.
    """
    global job_queue
    while True:
        attempt = job_queue.get()
        if not attempt['evaluated']:
            attid = attempt['attempt_id']
            tools.log('Checking attempt {}'.format(attid))
            commands = attempt['commands']
            out_list = attempt['out_list']
            attempt['status'] = []
            for command, out_expected in zip(commands, out_list):
                ran_without_error, out, err = __run_command__(command,
                                                              config.timeout)
                status = False if ran_without_error else None
                tools.log(ran_without_error, out, err, '-'*10)
                if ran_without_error:
                    status = all((i.strip() == j.strip())
                                 for i, j in zip(out, out_expected))
                attempt['status'].append(status)
                tools.log('Attempt {} status {}'.format(attid, status))
            attempt['evaluated'] = True
            tools.add_attempt_to_contest(attempt)
        else:
            tools.log('This attempt was in Job queue but checked')
            tools.log(attempt)
            tools.log('*'*50)


def __run_command__(command, timeout):
    "Run a given command with a set timeout"
    try:
        p = run(command, timeout=timeout,
                stderr=PIPE, stdout=PIPE,
                shell=True)
    except TimeoutExpired:
        result = False
        stdout, stderr = b'', b'Timeout'
    else:
        result = p.returncode == 0
        stdout, stderr = p.stdout, p.stderr
    return result, stdout.decode(), stderr.decode()


# -------------------------------------------------------------
job_queue = Queue()
# Start threads to check the code
for tid in range(config.n_threads_to_check_threads):
    t = Thread(target=__thread_worker__)
    t.start()
    config.code_checking_threads[tid] = t
# -------------------------------------------------------------


def submit_attempt(code, inp_list, out_list, wrap, attempt_id, user, qpk):
    global job_queue
    assert len(inp_list) == len(out_list), 'each inp needs an out'
    codepath = os.path.join(config.working_root, attempt_id)
    with open(codepath, 'w') as fl:
        fl.write(code)
    inp_paths, commands = [], []
    for inp, out in zip(inp_list, out_list):
        inpath = os.path.join(config.working_root, tools.random_id())
        inp_paths.append(inpath)
        with open(inpath, 'w') as fl:
            fl.write(inp)
        command = wrap.format(code=codepath, input=inpath)
        commands.append(command)
    tools.log('Adding attempt {} to job queue'.format(attempt_id))
    job_queue.put({'evaluated': False, 'attempt_id': attempt_id,
                   'inp_paths': inp_paths,
                   'code_path': codepath, 'commands': commands,
                   'out_list': out_list, 'user': user,
                   'qpk': qpk
                   })


def get_attempt_status(attempt_id):
    with tools.Contest() as contest:
        if attempt_id not in contest['attempts']:
            result, remark = None, 'The attempt has been sent to the judge'
        else:
            attempt = contest['attempts'][attempt_id]
            status = attempt['status']
            if any(i is None for i in status):
                remark = 'The program raised an error'
                result = False
            elif all(status):
                remark = 'Your Program Cleared {} tests'
                remark = remark.format(sum(1 for i in status if i))
                result = True
            elif not all(status):
                remark = 'Your Program did not clear some tests.'
                result = False
    return result, remark



import bottle
from openjudge import tools, config, judge


app = bottle.Bottle()


def jget(*keys):
    "Needs to be used like a, b, c = jget(x, y, z)"
    return [bottle.request.json[key] for key in keys]


@app.get('/')
def home():
    with tools.Contest() as contest:
        d = {'languages': list(sorted(list(contest['wrappers'].keys()))),
             'questions': list(sorted(list(contest['questions'].keys()),
                                      key=lambda x: int(x))),
             'intro': contest['intro']
             }
    return tools.render('home.html', d)


@app.get('/update_analysis')
def analytics_update():
    tools.update_analysis()
    return {'status': True}


@app.get('/analysis')
def analytics():
    d = {'available': config.analysis_available}
    return tools.render('analytics.html', d)


@app.get('/static/<path:path>')
def static_server(path):
    root = config.static_root
    return bottle.static_file(path, root=root)


# ---------------------------------------------API
@app.post('/login')
def login():
    u, p = jget('username', 'password')
    status, token = tools.login_user(u, p)
    return {'status': status, 'token': token}


@app.post('/logout')
def logout():
    token, = jget('token')
    return {'status': tools.logout_user(token)}


@app.post('/register')
def register():
    u, p = jget('username', 'password')
    status = tools.register_user(u, p)
    return {'status': status}


@app.post('/question')
def question_display():
    pk, = jget('question_pk')
    statement = 'This question does not exist yet.'
    with tools.Contest() as contest:
        pk = str(pk)
        q = {'statement': 'This question does not exist'}
        if pk.isdigit():
            if pk in contest['questions']:
                q = contest['questions'][pk]
                statement = q['statement']
    return {'statement': statement}


@app.post('/attempt')
def question_attempt():
    qpk, lang, code, token = jget('question', 'language', 'code', 'token')
    user = tools.get_user(token)
    message, attid = 'Unexpected Error', None
    if user is not None:
        if tools.attempt_is_ok(qpk, lang, code):
            i, o = tools.get_question_io(qpk)
            wrap = tools.get_wrap(lang)
            attid = tools.random_id()
            judge.submit_attempt(code, i, o, wrap, attid, user, qpk)
            message = 'Submitted'
        else:
            message = 'Question/Language does not exist'
    else:
        message = 'Please login'
    # -------------------------------------------
    return {'attempt': attid, 'message': message}


@app.post('/attempt/status')
def attempt_status():
    attid, = jget('attempt')
    status, message = judge.get_attempt_status(attid)
    return {'status': status, 'message': message}


@app.post('/user/score')
def user_score():
    user, = jget('user')
    score = tools.get_user_score(user)
    return {'score': score}


@app.post('/user/leader')
def user_list():
    users = tools.get_all_users()
    data = [(tools.get_user_score(u), u) for u in users]
    data.sort(key=lambda x: x[0], reverse=True)
    return {'leader': data}


@app.post('/user/details')
def user_details():
    token,  = jget('token')
    user = tools.get_user(token)['name']
    score = tools.get_user_score(user)
    return {'user': user, 'score': score}



import sys
from setuptools import setup
from openjudge import __version__
__version__ = list(map(str, __version__))

if sys.version_info < (3, 5):
    raise Exception('You require Python3.5 or above to run OpenJudge')

setup(name='openjudge',
      version='.'.join(__version__),
      description='LAN Programming Contest Judging Program',
      url='http://github.com/theSage21/openJudge',
      author='Arjoonn Sharma',
      author_email='arjoonn.94@gmail.com',
      license='MIT',
      packages=['openjudge'],
      include_package_data=True,
      install_requires=['bottle'],
      extras_require={'analysis': ['pandas', 'matplotlib']},
      entry_points={'console_scripts': ['openjudge=openjudge.cli:main']},
      package_data={'openjudge': ['templates/*', 'static/*']},
      keywords=['openjudge', 'lan', 'programming', 'programming', 'contest'],
      zip_safe=False)
